from hitchdb.tbls import TBLSConfig
from hitchdb.fixture import Fixture
from pathlib import Path
import json


class HitchDb:
    def __init__(self, tbls_json_path: Path):
        self._tbls_json_path = Path(tbls_json_path)

        assert self._tbls_json_path.exists()
        tbls_json = json.loads(self._tbls_json_path.read_text())
        assert tbls_json["driver"]["name"] == "postgres", (
            "HitchDb currently only works for postgres.\n"
            "Raise a ticket for mysql/mariadb/"
            "microsoft sql server/sqllite/bigquery/redshift/dynamodb"
        )

        self._tbls_config = TBLSConfig(tbls_json)

    def strictyaml_schema(self):
        return self._tbls_config.strictyaml_schema()

    def fixture(self, data: dict):
        assert isinstance(data, dict)
        return Fixture(
            data,
            self._tbls_config,
        )



from hitchdb.utils import sqlformat


INSERT_INTO = """
INSERT INTO {table_name} ({column_list})
VALUES
    {value_list}
"""


VALUE_LINE = """({pk}, {row}),\n    """


LAST_VALUE_LINE = """({pk}, {row});\n"""


class Fixture:
    def __init__(self, fixture_dict, tbls_config):
        self._fix_dict = fixture_dict
        self._tbls_config = tbls_config

    def _value_list(self, table, data_dict, column_list):
        tbl = self._tbls_config._tbl_dict[table]
        sql_text = ""
        items = list(data_dict.items())

        for pk, data in items[:-1]:
            sql_text += VALUE_LINE.format(
                pk=sqlformat(pk),
                row=", ".join([sqlformat(x) for x in tbl.row_list(data)]),
            )

        pk, data = items[-1]
        sql_text += LAST_VALUE_LINE.format(
            pk=sqlformat(pk), row=", ".join([sqlformat(x) for x in tbl.row_list(data)])
        )

        return sql_text

    def sql(self):
        sql_text = ""

        for table, data_dict in self._fix_dict.items():
            column_list = self._tbls_config.column_list(table)
            sql_text += INSERT_INTO.format(
                table_name=table,
                column_list=", ".join(column_list),
                value_list=self._value_list(table, data_dict, column_list),
            )

        return sql_text



from slugify import slugify
import re


def readable(name):
    """
    Formulate a readable equivalent of column and table
    names, whether they use camelCase or underscore_case.

    >>> readable("firstNameAndLastName")
    'first name and last name'

    >>> readable("first_name_and_last_name")
    'first name and last name'
    """
    return slugify(re.sub(re.compile("([a-z])([A-Z])"), r"\1 \2", name)).replace(
        "-", " "
    )


def sqlformat(scalar):
    """
    >>> sqlformat("hello")
    "'hello'"

    >>> sqlformat(45)
    '45'

    >>> sqlformat(45.3)
    '45.3'

    >>> sqlformat(True)
    'true'

    >>> sqlformat(False)
    'false'
    """
    if isinstance(scalar, str):
        return f"'{scalar}'"
    elif isinstance(scalar, bool):
        return "true" if scalar else "false"
    elif isinstance(scalar, int):
        return str(scalar)
    elif isinstance(scalar, float):
        return str(scalar)
    else:
        raise NotImplementedError(f"Unimplemented type: {type(scalar)}")


def sqlescape(sql):
    """
    Escape an SQL string.

    >>> sqlescape("O'Connor")
    "O''Connor"

    >>> sqlescape("\\n")
    '\\\\n'
    """
    return sql.translate(
        sql.maketrans(
            {
                "'": "''",
                "\0": "\\0",
                "\n": "\\n",
                '"': "",
                "\x09": "\\t",
                "\x1a": "\\z",
                "\x08": "\\b",
                "\r": "\\r",
                "\\": "\\\\",
                "%": "\\%",
            }
        )
    )
